﻿//---------------------------------------------------------------------------- 
//
//  Copyright (C) CSharp Labs.  All rights reserved.
// 
//  Permission is hereby granted, free of charge, to any person obtaining a copy
//  of this software and associated documentation files (the "Software"), to deal
//  in the Software without restriction, including without limitation the rights
//  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
//  copies of the Software, and to permit persons to whom the Software is
//  furnished to do so, subject to the following conditions:
// 
//  The above copyright notice and this permission notice shall be included in
//  all copies or substantial portions of the Software.
// 
//  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
//  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
//  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
//  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
//  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
//  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
//  THE SOFTWARE.
// 
// History
//  05/15/13    Created 
//  12/30/14    Fixed dst detection bugs.
//
// Remarks
//  Utilizes WebGrease (http://nuget.org/packages/WebGrease) to optimize
//  resulting JavaScript code.
//---------------------------------------------------------------------------

namespace System.Web.Util
{
    using global::Microsoft.Ajax.Utilities;
    using System.Collections.Generic;
    using System.Collections.ObjectModel;
    using System.Globalization;
    using System.Linq;
    using System.Text;

    /// <summary>
    /// Allows time zone detection through JavaScript.
    /// </summary>
    public class DynamicWebTimeZone
    {
        #region Fields
        /// <summary>
        /// Defines JavaScript to detect a user's time zone.
        /// </summary>
        public readonly string TimeZoneDetectorJavascript;

        /// <summary>
        /// Defines a collection of time zones for a combo box. Note the collection key is intended to be a drop down list item's value.
        /// </summary>
        private readonly ReadOnlyDictionary<string, string> TimeZones;
        #endregion

        #region Classes
        /// <summary>
        /// Defines time zone's id and display name.
        /// </summary>
        public class TimeZoneIdentifier
        {
            /// <summary>
            /// Defines the time zone id.
            /// </summary>
            public readonly string ID;
            /// <summary>
            /// Defines the time zone display name.
            /// </summary>
            public readonly string DisplayName;

            /// <summary>
            /// Initializes the TimeZoneIdentifier with an id and display name.
            /// </summary>
            /// <param name="id">The time zone id.</param>
            /// <param name="displayName">The time zone display name.</param>
            public TimeZoneIdentifier(string id, string displayName)
            {
                ID = id;
                DisplayName = displayName;
            }
        }

        /// <summary>
        /// TimeZoneData value type contains the time zone and day light savings time start date or null.
        /// </summary>
        private class TimeZoneData
        {
            /// <summary>
            /// Defines the time zone.
            /// </summary>
            public readonly TimeZoneInfo TimeZone;
            /// <summary>
            /// Defines the start time or null.
            /// </summary>
            public readonly DateTime? DSTStart;
            /// <summary>
            /// Determines if the time zone supports day light savings time.
            /// </summary>
            public readonly bool DayLightSavingsTime;
            /// <summary>
            /// Initializes the TimeZoneData with a time zone and dst value or null.
            /// </summary>
            /// <param name="timeZone">The time zone this represents.</param>
            /// <param name="dstStart">The day light savings time start time or null.</param>
            /// <param name="dst">Determines if the time zone supports day light savings time.</param>
            public TimeZoneData(TimeZoneInfo timeZone, DateTime? dstStart, bool dst)
            {
                TimeZone = timeZone;
                DSTStart = dstStart;
                DayLightSavingsTime = dst;
            }
        }
        #endregion

        #region Constructor
        /// <summary>
        /// Initializes the DynamicWebTimeZone class by reading all system time zones and creating the JavaScript time zone detector code.
        /// </summary>
        [System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Usage", "CA2214:DoNotCallOverridableMethodsInConstructors")]
        public DynamicWebTimeZone()
        {
            TimeZoneInfo[] zones = TimeZoneInfo.GetSystemTimeZones().ToArray(); //get all time zones
            int year = DateTime.UtcNow.Year; //the year to check day light savings time values for
            DateTime janUTC = new DateTime(year, 1, 1, 0, 0, 0, DateTimeKind.Utc); //jan 1st
            DateTime juneUTC = new DateTime(year, 6, 1, 0, 0, 0, DateTimeKind.Utc); //june 1st

            Dictionary<string, List<TimeZoneData>> timeZoneGroups = new Dictionary<string, List<TimeZoneData>>(); //time zone groups
            Dictionary<string, Dictionary<DateTime, TimeZoneData>> timeZoneDSTGroups = new Dictionary<string, Dictionary<DateTime, TimeZoneData>>(); //time zone and day light starting time groups
            Dictionary<string, string> timeZones = new Dictionary<string, string>(); //used to populate a combo box with data to match javascript data

            //Here we initialize preferred time zones to use instead of other identical zones:
            List<TimeZoneInfo> preferredZones = new List<TimeZoneInfo>();

            foreach (string zone in GetPreferredTimeZones())
                AddPreferredZone(preferredZones, zone);
            //test showing support for zones with dst -> AddPreferredZone(TimeZoneInfo.FindSystemTimeZoneById("Central European Standard Time"));

            // Construct a comparison objet for sorting day light savings time start dates:
            Comparison<TimeZoneData> timeZoneDataSorter = new Comparison<TimeZoneData>((first, second) =>
            {
                if (first.DSTStart.HasValue)
                {
                    if (second.DSTStart.HasValue)
                        return first.DSTStart.Value.CompareTo(second.DSTStart.Value);

                    return 1;
                }

                if (second.DSTStart.HasValue)
                    return -1;

                return 0;
            });

            //here we group up similar time zones:
            for (int i = 0; i < zones.Length; i++)
            {
                TimeZoneInfo zone = zones[i];

                //add zone details to cache for combo box:
                string zoneID = zone.Id;
                if (!timeZones.ContainsKey(zoneID))
                    timeZones.Add(zoneID, zone.DisplayName);

                //get the january and june utc offsets:
                int jan_UTC_Offset = -(int)(janUTC - TimeZoneInfo.ConvertTime(janUTC, zone)).TotalMinutes;
                int june_UTC_Offset = -(int)(juneUTC - TimeZoneInfo.ConvertTime(juneUTC, zone)).TotalMinutes;
                int jan_june_UTC_Diff = jan_UTC_Offset - june_UTC_Offset;

                string zone_id;
                if (jan_june_UTC_Diff < 0)
                    zone_id = String.Format("\"{0}|1\"", jan_UTC_Offset);
                else if (jan_june_UTC_Diff > 0)
                    zone_id = String.Format("\"{0}|2\"", june_UTC_Offset);
                else //no day light savings
                    zone_id = String.Format("\"{0}|0\"", jan_UTC_Offset);

                //get the time adjustments:
                TimeZoneInfo.AdjustmentRule[] rules = zone.GetAdjustmentRules();
                TimeZoneInfo.AdjustmentRule rule;
                bool dst = GetAdjustment(rules, year, out rule);

                List<TimeZoneData> lst;
                Dictionary<DateTime, TimeZoneData> dst_group;

                if (timeZoneGroups.ContainsKey(zone_id))
                {
                    lst = timeZoneGroups[zone_id];
                    dst_group = timeZoneDSTGroups[zone_id];
                }
                else
                {
                    lst = new List<TimeZoneData>();
                    timeZoneGroups.Add(zone_id, lst);
                    dst_group = new Dictionary<DateTime, TimeZoneData>();
                    timeZoneDSTGroups.Add(zone_id, dst_group);
                }

                if (dst) //all zones with day light savings time are supported with unique start times
                {
                    DateTime dstStart = GetStartDate(rule, year).AddHours(1); //offset by an hour for javascript date
                    bool isPreferred = false;

                    if (!dst_group.ContainsKey(dstStart) || (isPreferred = preferredZones.Contains(zone))) //check if this is a unique or preferred date
                    {
                        TimeZoneData data = new TimeZoneData(zone, dstStart, dst);

                        if (isPreferred)
                            lst.Remove(dst_group[dstStart]);

                        dst_group[dstStart] = data;
                        lst.Add(data);
                    }
                }
                else if (lst.Count == 0 || lst[0].DSTStart.HasValue) //we can only support a single zone for a specific UTC without day light savings time if no other null DST zones, insert at the front
                    lst.Insert(0, new TimeZoneData(zone, null, false));
                else if (preferredZones.Contains(zone)) //if preferred, replace the other zone
                    lst[0] = new TimeZoneData(zone, null, false);
            }

            StringBuilder tz_time_zones = new StringBuilder(); //time zones
            StringBuilder ambiguity_list = new StringBuilder(); //ambiguous time zones
            StringBuilder dst_start_dates = new StringBuilder(); //ambiguous time zone day light savings times
            List<string> wordLookup = SortAndCreateWordIndex(timeZoneDataSorter, timeZoneGroups); //create a word index to save bandwidth

            //here we initialize the javascript variables:
            foreach (var timeZone in timeZoneGroups)
            {
                List<TimeZoneData> timeZoneGroup = timeZone.Value;

                if (timeZoneGroup.Count > 0)
                {
                    TimeZoneData altZone = timeZoneGroup[0];

                    if (tz_time_zones.Length > 0)
                        tz_time_zones.Append(", ");

                    //print the time zone entry:
                    TimeSpan baseUtcOffset = altZone.TimeZone.BaseUtcOffset;

                    bool flag = baseUtcOffset.Hours < 0 || baseUtcOffset.Minutes < 0;

                    tz_time_zones.Append(timeZone.Key);
                    tz_time_zones.Append(": [\"");
                    tz_time_zones.Append(String.Format("{0}{1:D2}{2:D2}", (flag ? "-" : "+"), Math.Abs(baseUtcOffset.Hours), Math.Abs(baseUtcOffset.Minutes)));
                    tz_time_zones.Append("\", \"");
                    string packedId = ReplaceIDWords(wordLookup, altZone.TimeZone.Id);

                    tz_time_zones.Append(packedId);
                    tz_time_zones.Append("\", ");
                    tz_time_zones.Append(altZone.DayLightSavingsTime ? "!0" : "!1");
                    tz_time_zones.Append("]");

                    if (timeZoneGroup.Count > 1) //check for alternative entries
                    {
                        if (ambiguity_list.Length > 0)
                            ambiguity_list.Append(", ");

                        //print the zone id for the ambiguity list:
                        ambiguity_list.Append(String.Format("\"{0}\": [", packedId));

                        for (int i = 0; i < timeZoneGroup.Count; i++)
                        {
                            altZone = timeZoneGroup[i];

                            if (i > 0)
                            {
                                packedId = ReplaceIDWords(wordLookup, altZone.TimeZone.Id);
                                ambiguity_list.Append(", ");
                            }

                            ambiguity_list.Append(String.Format("\"{0}\"", packedId));

                            if (altZone.DSTStart.HasValue)
                            {
                                if (dst_start_dates.Length > 0)
                                    dst_start_dates.Append(", ");

                                DateTime startTime = altZone.DSTStart.Value;
                                dst_start_dates.Append(String.Format("\"{0}\": new Date({1}, {2}, {3}, {4}, {5}, {6}, {7})", packedId, startTime.Year, startTime.Month - 1, startTime.Day, startTime.Hour, startTime.Minute, startTime.Second, startTime.Millisecond));
                            }
                        }

                        ambiguity_list.Append("]");
                    }
                }
            }

            TimeZoneDetectorJavascript = MinifyJavaScriptTimeZoneDetectorCode(CreateJavaScriptTimeZoneDetectorCode(tz_time_zones, dst_start_dates, ambiguity_list, wordLookup));
            TimeZones = new ReadOnlyDictionary<string, string>(timeZones);
        }
        #endregion

        #region Methods
        /// <summary>
        /// Determines if the time zone id represents a valid time zone.
        /// </summary>
        /// <param name="id">The time zone identifier, which corresponds to the System.TimeZoneInfo.Id property.</param>
        /// <returns>true if the time zone is valid; otherwise, false.</returns>
        public bool ContainsTimeZone(string id)
        {
            return TimeZones.ContainsKey(id);
        }

        /// <summary>
        /// Gets time zone data suitable for a drop down list.
        /// </summary>
        /// <returns>TimeZoneIdentifier classes with time zone id and display names.</returns>
        public IEnumerable<DynamicWebTimeZone.TimeZoneIdentifier> GetTimeZones()
        {
            foreach (var zone in TimeZones)
                yield return new TimeZoneIdentifier(zone.Key, zone.Value);
        }

        /// <summary>
        /// There are some identical time zones, this method allows us to make a preference on which time zone
        /// is initially selected. Enter time zone ids that represent your target geographic areas.
        /// </summary>
        /// <returns></returns>
        protected virtual IEnumerable<string> GetPreferredTimeZones()
        {
            foreach (string zone in new string[] { "Russian Standard Time", "Tokyo Standard Time", "UTC" })
                yield return zone;
        }
        #endregion

        #region Static Helper Methods
        /// <summary>
        /// Uses WebGrease library to minify JavaScript.
        /// </summary>
        private static string MinifyJavaScriptTimeZoneDetectorCode(string javascript)
        {
            Minifier minifier = new Minifier();
            CodeSettings codeSettings = new CodeSettings
            {
                EvalTreatment = EvalTreatment.MakeImmediateSafe,
                PreserveImportantComments = false
            };

            return minifier.MinifyJavaScript(javascript, codeSettings);
        }

        /// <summary>
        /// Creates the JavaScript time zone detector code.
        /// </summary>
        private static string CreateJavaScriptTimeZoneDetectorCode(StringBuilder tz_time_zones, StringBuilder dst_start_dates, StringBuilder ambiguity_list, List<string> wordLookup)
        {
            //write javascript detector code:
            StringBuilder sb = new StringBuilder();

            sb.Append(@" (function (SystemWorks, undefined) {        SystemWorks.TimeZone = function () {
            var getOffset = function (dt) {
                dt = -dt.getTimezoneOffset();
                return dt !== null ? dt : 0;
            }, getFullYear = function(dt) {
                var year = dt.getFullYear();
                if (year < 1000)
                    year+=1900;
                return year;
            }, getJanuaryOffset = function (dt) {
                return getOffset(new Date(getFullYear(dt), 0, 1, 0, 0, 0, 0));
            }, getJuneOffset = function (dt) {
                return getOffset(new Date(getFullYear(dt), 5, 1, 0, 0, 0, 0));
            }, getPattern = function (dt) {
                var janUTC = getJanuaryOffset(dt), junUTC = getJuneOffset(dt), utcDiff = janUTC - junUTC;

                if (utcDiff < 0)");
            sb.Append("return janUTC + \"|1\"; else if (utcDiff > 0) return junUTC + \"|2\"; else return getOffset(dt) + \"|0\";  };");

            sb.Append(@"return {
                determine_timezone: function () {
                    var now = new Date(), pattern = getPattern(now);
                    return new SystemWorks.TimeZone.TimeZone(SystemWorks.TimeZone.timezones[pattern]);
                }, date_is_dst: function (dt) {
                    if (dt == null)
                        return false;
                    var utcOffset = dt.getMonth() > 5 ? getJuneOffset(dt) : getJanuaryOffset(dt), dt = getOffset(dt);
                    return utcOffset - dt !== 0;
                }
            }
        } ();

        SystemWorks.TimeZone.TimeZone = function () {
            var zone_name = null, is_dst = null, utcOffset = null, success, getTimeZone = function (zone) {
                if (success = !!zone)
                {
                utcOffset = zone[0];
                zone_name = zone[1];
                is_dst = zone[2];");
            sb.Append("if (typeof SystemWorks.TimeZone.ambiguity_list[zone_name] !== typeof undefined) {");
            sb.Append(@"for (var other = SystemWorks.TimeZone.ambiguity_list[zone_name], count = other.length, i = 0, other_zone_name = other[0]; i < count; i += 1) {
                        if (other_zone_name = other[i], SystemWorks.TimeZone.date_is_dst(SystemWorks.TimeZone.dst_start_dates[other_zone_name])) {
                            zone_name = other_zone_name;
                            break;
                        }
                    }
                }
                }
            };
            getTimeZone.prototype = { constructor: SystemWorks.TimeZone.TimeZone, success : function () {return success;}, name: function () { return success ? zone_name.replace(/%(\d+)/g, function(s, m1) { return SystemWorks.TimeZone.timeZone_Index[parseInt(m1)];}) : null; }, dst: function () { return is_dst; }, offset: function () { return utcOffset; } };
            return getTimeZone;
        } ();");

            sb.Append("SystemWorks.TimeZone.timezones = function () { return { ");
            sb.Append(tz_time_zones.ToString());
            sb.Append(" } } (); SystemWorks.TimeZone.dst_start_dates = function () { return { ");
            sb.Append(dst_start_dates.ToString());
            sb.Append(" } } (); SystemWorks.TimeZone.ambiguity_list = { ");
            sb.Append(ambiguity_list.ToString());
            sb.Append(" }; SystemWorks.TimeZone.timeZone_Index = [");

            if (wordLookup.Count > 0)
                sb.Append("\"" + string.Join("\",\"", wordLookup) + "\"");

            sb.Append("]; }(window.SystemWorks = window.SystemWorks || {}));");

            return sb.ToString();
        }


        /// <summary>
        /// This function gets the DateTime for a rule in a specific year.
        /// </summary>
        /// <param name="rule">The adjustment rule.</param>
        /// <param name="year">The year to apply the rule to.</param>
        /// <returns>The start date of the adjustment rule.</returns>
        private static DateTime GetStartDate(TimeZoneInfo.AdjustmentRule rule, int year)
        {
            TimeZoneInfo.TransitionTime transition = rule.DaylightTransitionStart;

            if (rule.DaylightTransitionStart.IsFixedDateRule)
                return new DateTime(year, transition.Month, transition.Day);
            else
            {
                // For non-fixed date rules, get local calendar
                System.Globalization.Calendar cal = CultureInfo.CurrentCulture.Calendar;
                // Get first day of week for transition
                // For example, the 3rd week starts no earlier than the 15th of the month
                int startOfWeek = transition.Week * 7 - 6;
                // What day of the week does the month start on?
                int firstDayOfWeek = (int)cal.GetDayOfWeek(new DateTime(year, transition.Month, 1));
                // Determine how much start date has to be adjusted
                int transitionDay;
                int changeDayOfWeek = (int)transition.DayOfWeek;

                if (firstDayOfWeek <= changeDayOfWeek)
                    transitionDay = startOfWeek + (changeDayOfWeek - firstDayOfWeek);
                else
                    transitionDay = startOfWeek + (7 - firstDayOfWeek + changeDayOfWeek);

                // Adjust for months with no fifth week
                if (transitionDay > cal.GetDaysInMonth(year, transition.Month))
                    transitionDay -= 7;

                DateTime timeOfDay = transition.TimeOfDay;

                return new DateTime(year, transition.Month, transitionDay, timeOfDay.Hour, timeOfDay.Minute, timeOfDay.Second);
            }
        }

        /// <summary>
        /// This method is intended to find TimeZoneInfo adjustment rules for a specific year.
        /// </summary>
        /// <param name="adjustments">A collection of adjustments.</param>
        /// <param name="year">The year to locate the adjustments for.</param>
        /// <param name="adj">out: the rule located.</param>
        /// <returns>true if an adjustment was located; otherwise, false.</returns>
        private static bool GetAdjustment(TimeZoneInfo.AdjustmentRule[] adjustments, int year, out TimeZoneInfo.AdjustmentRule adj)
        {
            // Iterate adjustment rules for time zone
            foreach (TimeZoneInfo.AdjustmentRule adjustment in adjustments)
            {
                // Determine if this adjustment rule covers year desired
                if (adjustment.DateStart.Year <= year && adjustment.DateEnd.Year >= year)
                {
                    adj = adjustment;
                    return true;
                }
            }

            adj = null;
            return false;
        }

        /// <summary>
        /// Adds a preferred timezones to use.
        /// </summary>
        /// <param name="preferredZones">A collection of TimeZoneInfo to prefer over duplicate zones.</param>
        /// <param name="id">The new timezone id to prefer.</param>
        private static void AddPreferredZone(List<TimeZoneInfo> preferredZones, string id)
        {
            try
            {
                preferredZones.Add(TimeZoneInfo.FindSystemTimeZoneById(id));
            }
            catch (TimeZoneNotFoundException)
            { }
        }

        /// <summary>
        /// This will split up words and tally the references each word appears in the final JavaScript code.
        /// </summary>
        /// <param name="id"></param>
        /// <param name="increment"></param>
        private static void AddToIndex(Dictionary<string, int> wordIndex, string id, int increment)
        {
            char[] chars = id.ToCharArray();

            for (int i = 0; i < chars.Length; i++)
                if (!char.IsLetter(chars[i]))
                    chars[i] = ' ';

            string[] words = (new string(chars)).Split(new char[] { ' ' }, StringSplitOptions.RemoveEmptyEntries);

            foreach (string word in words)
            {
                if (word.Length > 3)
                {
                    if (wordIndex.ContainsKey(word))
                        wordIndex[word] += increment;
                    else
                        wordIndex.Add(word, increment);
                }
            }
        }

        /// <summary>
        /// Sorts the time zone groups and returns a word lookup.
        /// </summary>
        /// <param name="timeZoneGroups">The time zone group lookup.</param>
        /// <returns>A collection of words to replace.</returns>
        private static List<string> SortAndCreateWordIndex(Comparison<TimeZoneData> sorter, Dictionary<string, List<TimeZoneData>> timeZoneGroups)
        {
            List<string> wordIndex = new List<string>();
            Dictionary<string, int> wordReferences = new Dictionary<string, int>();

            //next, build a dictionary of common words:
            foreach (var timeZone in timeZoneGroups)
            {
                List<TimeZoneData> timeZoneGroup = timeZone.Value;

                if (timeZoneGroup.Count > 0)
                {
                    //sort the groups by the DST start:
                    timeZoneGroup.Sort(sorter);

                    TimeZoneData altZone = timeZoneGroup[0];
                    if (timeZoneGroup.Count > 1) //check for alternative entries
                    {
                        AddToIndex(wordReferences, altZone.TimeZone.Id, 4);

                        for (int i = 1; i < timeZoneGroup.Count; i++)
                            AddToIndex(wordReferences, timeZoneGroup[i].TimeZone.Id, 2);
                    }
                    else
                        AddToIndex(wordReferences, altZone.TimeZone.Id, 1);
                }
            }

            foreach (var item in wordReferences)
            {
                if (item.Value > 2)
                    wordIndex.Add(item.Key);
            }

            return wordIndex;
        }

        /// <summary>
        /// Replace any words in the id from the index.
        /// </summary>
        /// <param name="wordLookup">The word lookup.</param>
        /// <param name="id">The time zone id.</param>
        /// <returns>A new id suitable to compact the JavaScript code.</returns>
        private static string ReplaceIDWords(List<string> wordLookup, string id)
        {
            for (int i = 0; i < wordLookup.Count; i++)
                id = id.Replace(wordLookup[i], "%" + i.ToString());

            return id;
        }
        #endregion
    }
}